; CALCULO DE PI Y LOGARITMO NATURAL DE 2 POR DARIO A. ALPERN
; USANDO DPMI (DOS PROTECTED MODE INTERFACE)
;
; Formulas utilizadas:
; ===================
;
; Calculo de pi:
;
;                   1            1            1             13753  
; pi = 4  6 arc tg  + arc tg  + arc tg  + arc tg  
;                   8           32           128          63664921 
;
;
; Calculo de ln 2:
;
;                    1             1             1529
; ln 2 =  5 arg tgh  + arg tgh  + arg tgh 
;                    8            16            670751
;
;
; El calculo del arco tangente se realiza mediante la formula de Taylor:
;
;                  3     5     7     9
;                 x     x     x     x
; arc tg x = x -  +  -  +  - ...
;                 3     5     7     9
;
; El calculo del argumento de la tangente hiperbolica se realiza mediante la
; formula de Taylor:
;
;                   3     5     7     9
;                  x     x     x     x
; arg tgh x = x +  +  +  +  + ...
;                  3     5     7     9
;
;
; Estas formulas se deducen teniendo en cuenta que:
;
;                                                          x + y
;   pi = 4 arc tg 1     y   arc tg x + arc tg y = arc tg  ;
;                                                         1 - xy
;
;                 3                                           x + y
; ln 2 = arg tgh     y   arg tgh x + arg tgh y = arg tgh  .
;                 5                                          1 + xy
;
; En el caso de procesadores de 16 bits es muy costoso en tiempo dividir por
; 63664921 o por 670751, por lo que se divide por sus factores (mediante dos
; divisiones). Las factorizaciones son: 63664921 = 1217 x 52313 y
; 670751 = 347 x 1933.
;
;----------------------------------------------------------------------------
;
; Durante el calculo de PI o LN 2, se utilizan tres buffers que contienen
; diferentes numeros en binario: el primero (que comienza en "buffer")
; contiene la suma parcial de los terminos; en el segundo (que comienza en
; "buffer + 16384") se calculan las potencias que intervienen en los
; diferentes terminos; mientras que en el ultimo (que comienza en
; "buffer + 32768") se calcula un termino que luego se sumara o restara del
; que comienza en "buffer".
;
;----------------------------------------------------------------------------
;
; Definicion de constantes y variables inicializadas del programa.
;
.386
eloop    macro destino            ;Macro para hacer loop en ECX.
         db 67h                   ;Prefijo de longitud de direccion.
         loop destino
         endm

estruc_modo_real struc        ;Estructura para pasar parametros a modo real.
reg_EDI  dd ?                 ;Imagen registro EDI.
reg_ESI  dd ?                 ;Imagen registro ESI.
reg_EBP  dd ?                 ;Imagen registro EBP.
reserved dd ?
reg_EBX  dd ?                 ;Imagen registro EBX.
reg_EDX  dd ?                 ;Imagen registro EDX.
reg_ECX  dd ?                 ;Imagen registro ECX.
reg_EAX  dd ?                 ;Imagen registro EAX.
reg_Flags  dw ?               ;Imagen registro de indicadores.
reg_ES   dw ?                 ;Imagen registro ES modo real.
reg_DS   dw ?                 ;Imagen registro DS modo real.
reg_FS   dw ?                 ;Imagen registro FS modo real.
reg_GS   dw ?                 ;Imagen registro GS modo real.
reg_IP   dw ?                 ;Imagen registro IP modo real.
reg_CS   dw ?                 ;Imagen registro CS modo real.
reg_SP   dw ?                 ;Imagen registro SP modo real.
reg_SS   dw ?                 ;Imagen registro SS modo real.
estruc_modo_real ends

HANDLE_STDOUT      equ 1      ;Handle Standard Output (DOS).
HANDLE_STDERR      equ 2      ;Handle Standard Error (DOS).

LN2                equ 1
E                  equ 2
Formato            equ 4
MostrarCantDigitos equ 8

ASCII_CR           equ 13     ;Mover el cursor al extremo izquierdo.
ASCII_LF           equ 10     ;Mover el cursor hacia abajo.
port6845           equ 3D4h   ;Port del controlador de video.

codigo  segment use16
        assume cs:codigo,ds:codigo
        org 100h            ;Offset de arranque del programa en formato .COM
comienzo:jmp Inicio         ;Salto al inicio del programa.

divisor         dd 1        ;Divisor de cada termino. Va tomando los valores
                            ;1, 3, 5,... como se aprecia en las formulas.

digitos         dd 0        ;Cantidad de digitos que posee la aproximacion que
                            ;se desea mostrar.

PMode_entry     label dword ;Direccion punto de entrada a modo protegido.
PMode_entry_offs dw 0       ;Offset.
PMode_entry_segm dw 0       ;Segmento.

CantDWords      dd 0        ;Cantidad de DWords (32 bits) correspondientes a
                            ;la cantidad de digitos pedidos.

UltDWord        dd    0
PrimDWord       dd    0     ;Palabra mas significativa distinta de cero.
                            ;Al no tener que procesar las palabras que valen
                            ;cero se duplica la velocidad.

Numerador       dd 13753*13753 ;Numerador (32 bits) del argumento del ultimo
                            ;arc tg/arg tgh. Esta inicializado para el calculo
                            ;de PI. En el caso del calculo de ln 2 este valor
                            ;se sobreescribe con 1529*1529 al principio del
                            ;programa. Ver formulas arriba.

Denominador     dd 63664921 ;Denominador (32 bits) del argumento del ultimo
                            ;arc tg/arg tgh. Se inicializa al principio del
                            ;programa con 63664921 (para PI) o 670751
                            ;(para ln 2). Ver formulas arriba.

; Multiplicadores para averiguar la cantidad de DWords W necesarios para
; calcular n digitos en base b: W = 2^32*log(b)/log(2^32)
Multiplicadores label dword
                dd 134217728    ; Base 2
                dd 212730066    ; Base 3
                dd 268435456    ; Base 4
                dd 311643913    ; Base 5
                dd 346947794    ; Base 6
                dd 376796799    ; Base 7
                dd 402653184    ; Base 8
                dd 425460132    ; Base 9
                dd 445861641    ; Base 10
                dd 464317052    ; Base 11
                dd 481165522    ; Base 12
                dd 496664612    ; Base 13
                dd 511014527    ; Base 14
                dd 524373979    ; Base 15
                dd 536870912    ; Base 16
                dd 548609976    ; Base 17
                dd 559677860    ; Base 18
                dd 570147180    ; Base 19
                dd 580079369    ; Base 20
                dd 589526865    ; Base 21
                dd 598534780    ; Base 22
                dd 607142208    ; Base 23
                dd 615383250    ; Base 24
                dd 623287827    ; Base 25
                dd 630882340    ; Base 26
                dd 638190197    ; Base 27
                dd 645232255    ; Base 28
                dd 652027172    ; Base 29
                dd 658591707    ; Base 30
                dd 664940973    ; Base 31
                dd 671088640    ; Base 32
                dd 677047118    ; Base 33
                dd 682827704    ; Base 34
                dd 688440713    ; Base 35
                dd 693895588    ; Base 36

MaxPotencia       dd 0      ;Maxima potencia para la base pedida que entre
                            ;en un doubleword.
                            ;El valor es: Base^(int(log (2^32)/log Base))
LogMaxPotencia    dd 0      ;Exponente para la potencia arriba indicada.
                            ;Cantidad de digitos por DWord.
                            ;El valor es: int(log (2^32)/log Base)) *
                            ;* 2^32*log Base/log(2^32)
SumaLogPotencia   dd 0      ;Suma de LogMaxPotencia para determinar la
                            ;cantidad de digitos durante la conversion.
Exponente         dd 0      ;Indica la cantidad de digitos que se convieten
                            ;por vez.
HandleMemoriaBajo dw 0      ;Handle de memoria que entrega
HandleMemoriaAlto dw 0      ;el servidor de DPMI.

PunteroSubrPotenciasDe2 dw PotBinParaPi
                            ;Puntero a la subrutina para calcular las potencias
                            ;binarias. Esta inicializada para el calculo de PI.
                            ;En el caso del calculo de ln 2 este valor se
                            ;sobreescribe con el correspondiente para ln 2.

handle          dw 1        ;Handle correspondiente al dispositivo estandar
                            ;de salida (la pantalla por defecto). Se
                            ;sobreescribe con el handle del archivo, si se
                            ;especifica.

Indicadores     db 0        ;Sirve para saber si la linea de comando contiene
                            ;los "switches" /pi o /ln2:
                            ;Bit 0 = 1: Calculo de ln 2.
                            ;Bit 0 = 0: Calculo de pi.
                            ;Bit 1 = 1: Salida formateada.
                            ;Bit 1 = 0: Salida sin formatear.

bufferdos       db "3,",60 dup (0) ;Buffer para la conversion de binario a
                                   ;decimal.

BufferNumeroFormateado db "3,",75 dup (" "),ASCII_CR,ASCII_LF
                            ;Buffer de una linea para obtener la salida
                            ;formateada.

PosicionDigitoEnBuffer dw BufferNumeroFormateado + 2
                            ;Puntero dentro del buffer de una linea.

InformacionDeFormateado dw 0      ;MSB = Numero de grupo (0-9).
                                  ;LSB = Digito dentro del grupo (1-5).

DigitosMostrados        dd 0

Selector dw 0

Iniciales     db "dhms"
Dias          db 0
Horas         db 0
Minutos       db 0
Segundos      db 0
FraccSegundo  dd 0

OldInt8Handler label dword
OldInt8Offset  dw 0
OldInt8Segment dw 0

MostrarEnInt8  db 1
PunteroBase    dw 0
Base           db 10


CursorVideo    label dword
CursorVideoOffset  dw 0
CursorVideoSegment dw 0B800h

CopyRight db "Calculo de PI, ln 2 y e. Hecho por Dario Alejandro Alpern."
        db ASCII_CR,ASCII_LF
LongitudCopyRight equ $ - CopyRight
Texto   db ASCII_CR,ASCII_LF
        db "Uso del programa: PIDPMI cantidad_de_decimales [/F] [/LN2] [/E] "
        db "[/B:base]",ASCII_CR,ASCII_LF," [arch_destino]",ASCII_CR,ASCII_LF
        db "   /F: Muestra la salida formateada para editores de texto"
        db ASCII_CR,ASCII_LF
        db "   /LN2: Muestra el valor del logaritmo natural de 2 en vez de pi"
        db ASCII_CR,ASCII_LF
        db "   /E: Muestra el valor del numero e en vez de pi"
        db ASCII_CR,ASCII_LF
        db "   /B: Muestra el resultado en base 2 a 36"
        db ASCII_CR,ASCII_LF
        db "   arch_destino: Archivo en que se almacenara el numero."
        db ASCII_CR,ASCII_LF
        db "                 Si no se especifica, es la pantalla (STDOUT)."
        db ASCII_CR,ASCII_LF
        db ASCII_CR,ASCII_LF
LongitudTexto equ $ - Texto
NoHayDPMI db "No hay servidor de DPMI.",ASCII_CR,ASCII_LF
LongitudNoHayDPMI equ $ - NoHayDPMI
Servidor16Bits db "El servidor de DPMI no soporta aplicaciones de 32 bits"
               db ASCII_CR,ASCII_LF
LongitudServidor16Bits equ $ - Servidor16Bits
NoHayMemoriaPServidor db "No hay memoria local para el servidor de DPMI."
               db ASCII_CR,ASCII_LF
LongitudNoHayMemoriaPServidor equ $ - NoHayMemoriaPServidor
NoSePuedePasarAProtegido db "No se puede pasar a modo protegido."
               db ASCII_CR,ASCII_LF
LongitudNoSePuedePasarAProtegido equ $ - NoSePuedePasarAProtegido
NoSeReservoSelector db "No se puede reservar un selector."
               db ASCII_CR,ASCII_LF
LongitudNoSeReservoSelector equ $ - NoSeReservoSelector
MemoriaLlena db ASCII_CR,ASCII_LF
        db "No hay suficiente memoria para procesar tantos "
        db "decimales.",ASCII_CR,ASCII_LF
LongitudMemoriaLlena equ $ - MemoriaLlena
TextoSegundaLinea db "Digitos hallados: "
LongitudTextoSegundaLinea equ $ - TextoSegundaLinea
TiempoTotal db ASCII_CR,ASCII_LF,"Tiempo total: "
LongitudTiempoTotal equ $ - TiempoTotal
ModoReal estruc_modo_real <0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0>
;=============================================================================
; Mostrar el "Copyright"
;=============================================================================
Inicio: mov dx,offset CopyRight ;Mostrar en pantalla el "copyright".
        mov cx,LongitudCopyRight
        mov bx,HANDLE_STDERR        ;Handle STDERR.
        mov ah,40h
        int 21h
;================================================================;
; Determinacion de existencia de DPMI y pasaje a modo protegido  ;
;================================================================;
VerificarDPMI:
        mov ModoReal.reg_DS,ds
        pushf
        pop ax
        and ax,0FFFh
        mov ModoReal.reg_Flags,ax
        mov ax,1687h            ;Ver si hay un servidor de DPMI instalado en
        int 2Fh                 ;memoria.
        test ax,ax
        jz short InicializarDPMI ;Saltar si hay servidor de DPMI.
        mov dx,offset NoHayDPMI
        mov cx,LongitudNoHayDPMI ;Mostrar que no hay servidor de DPMI.
error_en_real:
        mov bx,HANDLE_STDERR    ;Handle STDERR.
        mov ah,40h
        int 21h
        mov ax,4c01h            ;Fin del programa con codigo de error 1.
        int 21h
InicializarDPMI:
        mov PMode_entry_segm,es ;Preservar el punto de entrada del servidor
        mov PMode_entry_offs,di ;de DPMI para entrar a modo protegido.
        test bl,1               ;El servidor soporta aplicaciones de 32 bits?
        mov dx,offset Servidor16Bits
        mov cx,LongitudServidor16Bits
        jz short error_en_real  ;Saltar si no es asi.
        test si,si              ;El servidor necesita memoria local?
        jz short ir_a_modo_protegido ;Saltar si no es asi.
        push ds
        pop es                  ;ES = Segmento del programa .COM
        push si                 ;Preservar la cantidad de bytes que necesita
                                ;el servidor de DPMI.
        mov ah,4Ah              ;Achicar al minimo el bloque de memoria que
        mov bx,offset fin_codigo[15] ;incluye el programa .COM
        shr bx,4
        int 21h
        pop bx
        mov ah,48h
        int 21h                 ;Reservar memoria para el servidor de DPMI.
        mov cx,LongitudNoHayMemoriaPServidor
        mov dx,offset NoHayMemoriaPServidor
        jc short error_en_real  ;Saltar si no se pudo reservar memoria.
        mov es,ax               ;Apunta al area de datos del servidor de DPMI.
ir_a_modo_protegido:
        mov ax,1                ;Indicar que la aplicacion es de 32 bits.
        call PMode_entry        ;Pasar a modo protegido.
        mov cx,LongitudNoSePuedePasarAProtegido
        mov dx,offset NoSePuedePasarAProtegido
        jc short error_en_real
        push ds
        pop es                  ;Volver a poner ES = DS.
        xor ax,ax
        mov cx,1                ;Reservar un descriptor.
        int 31h
        mov cx,LongitudNoSeReservoSelector
        mov dx,offset NoSeReservoSelector
        jc short error_modo_prot
        mov Selector,ax         ;Preservar el selector correspondiente.
        mov bx,ax
        mov ax,0009h
        mov cx,40F2h
        int 31h                 ;Poner el byte de derechos de acceso.
HallarCantDecimales:
        mov bx,80h              ;Apuntar al byte anterior del principio de
                                ;la linea de comandos.
        mov esi,10              ;ESI = Multiplicador.
        xor eax,eax             ;Inicializar numero a formar a cero.
        xor ecx,ecx             ;Inicializar digito a cero.
BusquedaDeDigito:
        inc bx                  ;Apuntar al siguiente byte.
        mov al,[bx]             ;Obtener el caracter de la linea de comandos.
        cmp al,ASCII_CR         ;Se acabo la linea?
        je short error          ;Saltar a error si es asi.
        cmp al," "              ;Es un espacio o caracter no imprimible?
        jbe BusquedaDeDigito    ;Saltear dicho caracter.
        sub al,"0"              ;Ver si el caracter esta entre '0' y '9'.
        cmp al,9
        jbe short NumeroLineaDeComando ;Saltar si esta dentro del rango.
                                ;Caso contrario mostrar el error ya
                                ;que la linea debe comenzar con un numero.
error:
        mov dx,offset Texto     ;Mostrar texto indicando el formato de la
        mov cx,LongitudTexto    ;linea de comando.
error_modo_prot:
        mov bx,HANDLE_STDERR    ;Handle STDERR.
        mov ah,40h
        call dos
        mov ax,4c01h            ;Fin del programa con codigo de error 1.
        int 21h
NuevoDigito:
        mul esi                 ;Multiplicar numero a formar por diez.
        jc NoHaySuficienteMemoria ;Saltar si el resultado es muy grande.
        add eax,ecx             ;Sumar el digito hallado en la linea de comando
        jc NoHaySuficienteMemoria ;Saltar si el resultado es muy grande.
NumeroLineaDeComando:
        inc bx                  ;Apuntar al siguiente byte de la linea.
        mov cl,[bx]             ;Obtener el caracter.
        sub cl,"0"              ;Ver si esta entre '0' y '9'.
        cmp cl,9
        jbe NuevoDigito         ;Sumar digito y repetir si esta en el rango.
        mov digitos,eax         ;Almacenar cantidad de digitos en variable.
;
; Ver si en la linea de comandos aparecen /F o /LN2. En el caso que existan,
; se ponen a uno ciertos bits de la variable Indicadores. Para /F se pone a
; 1 el bit Formato, mientras que para /LN2 se pone a 1 el bit LN2.
;
        call switch             ;Encender indicadores segun los switches
                                ;encontrados en la linea de comandos.
        xor ecx,ecx
        mov cl,Base
        mov di,cx               ;Obtener la base.
        shl di,2
        add di,offset Multiplicadores[-8] ;Apuntar al logaritmo.
        mov PunteroBase,di
        mov si,-1               ;Inicializar exponente.
        mov eax,1               ;Inicializar potencia.
CicloHallarPotencia:
        inc si                  ;Actualizar exponente.
        mul ecx
        jnc CicloHallarPotencia
        mov word ptr Exponente,si
        div ecx
        mov MaxPotencia,eax     ;Almacenar la maxima potencia.
        mov eax,Exponente
        mul dword ptr [di]
        mov LogMaxPotencia,eax
HallarCantDWords:
        mov eax,Digitos         ;EAX = Cantidad de digitos a calcular.
        mov bx,PunteroBase      ;Obtener el puntero al multiplicador que
                                ;corresponde a la base.
        mul dword ptr [bx]      ;EDX = Cantidad de DWords necesarias.
        add edx,3               ;Reservar 3 DWords mas por el redondeo.
        mov CantDWords,edx      ;Almacenar la cantidad de DWords necesarias.
        shl edx,2               ;Convertir a cantidad de bytes.
        lea edx,[edx * 2 + edx] ;Reservar espacio para 3 buffers.
        or dx,0FFFh             ;Calcular el limite para pagina completa.
        shld ecx,edx,16         ;CX = Parte alta del limite.
        push cx                 ;Preservar la parte alta del limite.
        push dx                 ;Preservar la parte baja del limite.
        mov bx,Selector
        mov ax,0008h
        int 31h                 ;Poner el limite del segmento.
        pop cx                  ;Restaurar la parte baja del limite.
        pop bx                  ;Restaurar la parte alta del limite.
        add cx,1                ;Se debe reservar un byte mas que el limite.
        adc bx,0
        mov ax,0501h            ;Reservar BX:CX bytes de memoria lineal.
        int 31h                 ;BX:CX = Direccion lineal del bloque.
                                ;SI:DI = Handle de memoria.
        jnc short LosDigitosEntran ;Saltar si no hubo errores.
NoHaySuficienteMemoria:
        mov dx,offset MemoriaLlena ;Indicar la condicion de error.
        mov cx,LongitudMemoriaLlena
        mov ah,40h
        call dos
        mov ax,4C02h            ;Saltar si no es asi.
        int 21h
LosDigitosEntran:
        mov HandleMemoriaBajo,di  ;Preservar el handle de memoria ubicado
        mov HandleMemoriaAlto,si  ;en SI:DI
        mov dx,cx               ;DX = Parte baja de la base del segmento.
        mov cx,bx               ;CX = Parte alta de la base del segmento.
        mov ax,0007h
        mov bx,Selector
        int 31h                 ;Llenar la base del segmento.
InicializarPrimerTermino:
        test Indicadores,E
        jz NoEsCalculoE
;------------------------------------------------------------------------------
; Calculo del numero e:
;
        mov bufferdos,"2"       ;Parte entera de e.
        call PonerVectorInterrupcion
        mov ebx,CantDWords
        mov ecx,ebx             ;Obtener cantidad de DWords del numero.
        shl ecx,1               ;Obtener cantidad de DWords a poner a cero.
        xor edi,edi             ;Apuntar al principio del numero ubicado en
                                ;memoria pedida al servidor de DPMI.
        xor eax,eax             ;Poner ambos numeros a cero.
        rep stos dword ptr es:[edi]
        inc divisor                    ;Divisor = 2.
        mov byte ptr es:[ebx*4+3],80h  ;Poner a 0,5 el termino.
        mov byte ptr es:[3],80h        ;Poner a 2,5 la suma.
CicloPrincE:
        mov ecx,CantDWords             ;ECX = Cantidad de DWords del numero.
        inc divisor                    ;Siguiente divisor.
        mov edi,ecx                    ;EDI = Cantidad de DWords del numero.
        shl edi,2                      ;EDI = Puntero al termino.
        mov eax,PrimDWord              ;EAX = Numero de palabra mas
                                       ;significativa distinta de ceros.
        sub ecx,eax                    ;ECX = Cantidad de palabras a procesar.
        jbe BinarioADecimal            ;Saltar si es <= 0.
        lea edi,[edi+eax*4]            ;EDI = Puntero a la palabra mas
                                       ;significativa distinta de ceros.
        mov esi,divisor                ;ESI = Divisor en la formula de e.
        xor edx,edx                    ;EDX = Resto de la division.
        mov ebx,edi                    ;EBX = Puntero al MSD.
        cld                            ;Procesar en direcciones ascendentes
                                       ;(para realizar una division se deben
                                       ;tomar primero los DWords mas signif.)
        push ecx                       ;Preservar la cantidad de palabras a
                                       ;procesar.
        shr ecx,1
        jnc short DivVariableE
        mov eax,es:[edi]               ;Obtener el DWord del dividendo.
        div esi                        ;Realizar la primera division.
        stos dword ptr es:[edi]        ;Almacenar el cociente.
DivVariableE:
        jecxz FinDivVariableE          ;Saltar si se termino la division.
CicloDivVariableE:
        mov eax,es:[edi]               ;Obtener el DWord del dividendo.
        div esi                        ;Realizar la primera division.
        stos dword ptr es:[edi]        ;Almacenar el cociente.
        mov eax,es:[edi]               ;Obtener el DWord del dividendo.
        div esi                        ;Realizar la primera division.
        stos dword ptr es:[edi]        ;Almacenar el cociente.
        eloop CicloDivVariableE        ;Cerrar el ciclo.
FinDivVariableE:
        cmp dword ptr es:[ebx],1       ;El MSD es cero?
        adc PrimDWord,ecx              ;El MSD es el siguiente si lo era.
        std                            ;Procesar en direcciones descendentes
                                       ;(para realizar una suma se deben
                                       ;tomar primero los DWords menos
                                       ;significativos).
        lea esi,[edi-4]                ;Apuntar al LSD del termino.
        mov eax,CantDWords             ;EAX = Cantidad de DWords del numero.
        shl eax,2                      ;EAX = Cantidad de bytes del numero.
        mov edi,esi
        sub edi,eax                    ;EDI = Puntero al LSD de la suma.
        pop ecx                        ;Restaurar la cantidad de palabras a
                                       ;procesar.
        shr ecx,1
        jnc short SumaE                ;Saltar si la cant. de DWords era par.
        lods dword ptr es:[esi]        ;Obtener DWord del termino.
        adc eax,dword ptr es:[edi]     ;Sumarselo a la suma.
        stos dword ptr es:[edi]        ;Guardarlo en la suma.
SumaE:
        jecxz FinSumaE                 ;Saltar si se termino el sumando.
CicloSumaE:
        lods dword ptr es:[esi]        ;Obtener DWord del termino.
        adc eax,dword ptr es:[edi]     ;Sumarselo a la suma.
        stos dword ptr es:[edi]        ;Guardarlo en la suma.
        lods dword ptr es:[esi]        ;Obtener DWord del termino.
        adc eax,dword ptr es:[edi]     ;Sumarselo a la suma.
        stos dword ptr es:[edi]        ;Guardarlo en la suma.
        eloop CicloSumaE               ;Cerrar el ciclo.
FinSumaE:
        jnc CicloPrincE                ;Saltar si la suma se termino.
CicloSumaCarryE:
        inc dword ptr es:[edi]         ;Incrementar el siguiente DWord mas
                                       ;significativo.
        lea edi,[edi-4]                ;Apuntar al siguiente DWord mas
                                       ;significativo.
        jnz CicloPrincE                ;Saltar si se acabo la suma.
        jmp CicloSumaCarryE            ;Cerrar el ciclo.
;------------------------------------------------------------------------------
NoEsCalculoE:
        mov eax,13753*4           ;Primer numerador para PI.
        mov Numerador,13753*13753 ;Numerador para PI.
        mov Denominador,63664921  ;Denominador para PI.
        test Indicadores,LN2      ;Ver si se esta calculando PI o LN 2.
        jz short FinInicializacionCtes  ;Saltar si se esta calculando PI.
        mov PunteroSubrPotenciasDe2,offset PotBinParaLN2
                                  ;Subrutina a llamar cuando se calculen las
                                  ;potencias de 2.
        mov eax,1529
        mov Numerador,1529*1529  ;Numerador para LN 2.
        mov Denominador,670751   ;Denominador para LN 2.
        mov bufferdos,"0"        ;Indicar que el numero comienza con cero.
FinInicializacionCtes:
        push eax                 ;Preservar el primer numerador.
        call PonerVectorInterrupcion
        pop edx                 ;Obtener el primer numerador.
        mov ecx,CantDWords      ;Obtener cantidad de DWords del numero.
        lea esi,[ecx*4]         ;Convertir a bytes.
        xor edi,edi             ;Apuntar al principio del numero ubicado en
                                ;memoria pedida al servidor de DPMI.
        mov es,Selector         ;El segmento de datos apunta a dicha memoria.
        mov ebx,Denominador     ;Denominador del primer termino.
        cld                     ;Se procesa en direcciones crecientes.
CicloCalcPrimerTermino:
        xor eax,eax             ;La parte decimal del denominador es cero
                                ;en el primer termino.
        div ebx                 ;Generar DWord del cociente.
        mov es:[edi+esi],eax    ;Almacenarlo en el segundo numero.
        stos dword ptr es:[edi] ;Almacenarlo en el primer numero.
        eloop CicloCalcPrimerTermino ;Cerrar el ciclo.
        mov al,256*4/32+256*4/128    ;Agregar potencias de 2 para Pi.
        test Indicadores,LN2         ;Se esta calculando LN 2 ?
        jz short PonerPrimeraPot2    ;Saltar si no es asi.
        mov al,256*5/8+256*1/16      ;Agregar potencias de 2 para LN 2.
PonerPrimeraPot2:
        mov es:[3],al              ;Almacenarlo en el primer numero.
        mov es:[esi+3],al          ;Almacenarlo en el segundo numero.
CicloPrincipal:
        call NuevoTermino          ;Hallar el nuevo termino.
        test Indicadores,LN2       ;Se desea calcular LN 2?
        jnz short CicloSumaDelTermino ;Los arg tgh solo contienen sumas.
CicloRestaDelTermino:
        mov eax,es:[edi]           ;Obtener DWord de la suma.
        sbb eax,es:[esi]           ;Restarle el Dword del sustraendo.
        stos dword ptr es:[edi]    ;Guardar el DWord de la resta.
        lea esi,[esi-4]            ;Apuntar al siguiente DWord.
        eloop CicloRestaDelTermino ;Cerrar el ciclo.
        jnc short TerminoParaSuma  ;Saltar si se termino la resta.
CicloSeguirRestando:
        sbb dword ptr es:[edi],ecx ;Actualizar el DWord.
        lea edi,[edi-4]            ;Actualizar el puntero.
        jc CicloSeguirRestando     ;Saltar si no se acabo la resta,
TerminoParaSuma:
        call NuevoTermino          ;Hallar el nuevo termino.
CicloSumaDelTermino:
        lods dword ptr es:[esi]    ;Obtener el DWord del termino.
        adc eax,es:[edi]           ;Sumarle el DWord de la suma.
        stos dword ptr es:[edi]    ;Guardar el DWord de la suma.
        eloop CicloSumaDelTermino  ;Cerrar el ciclo.
        jnc CicloPrincipal         ;Saltar si la suma no se completo.
CicloSeguirSumando:
        adc dword ptr es:[edi],ecx ;Actualizar el DWord.
        lea edi,[edi-4]            ;Actualizar el puntero.
        jc CicloSeguirSumando      ;Saltar si no se acabo la suma.
        jmp CicloPrincipal         ;Cerrar el ciclo principal.
;--------------------------------------------------------------------------
; Subrutina PotBinParaLN2: Pone o saca las potencias binarias de 2 para el
; calculo del logaritmo natural de 2.
; Mediante un "xor", pone o saca las potencias de dos que figuran en la
; formula que figura al principio del programa.
;
PotBinParaLN2:
        mov edi,divisor
        lea edi,[edi*2+edi-3]
        mov cx,di
        mov eax,80000000h
        ror eax,cl
        shr edi,3
        and di,0FFFCh
        mov ecx,CantDWords
        xor es:[edi+ecx*4],eax
        mov eax,3
        mul divisor
        call CambiarBit
        mov eax,4
        mul divisor
        jmp short CambiarBit
;----------------------------------------------------------------------------
; Subrutina PotBinParaPI: Pone o saca las potencias binarias de 2 para el
; calculo de Pi.
; Mediante un "xor", pone o saca las potencias de dos que figuran en la
; formula que figura al principio del programa.
;
PotBinParaPi:
;
; Hay que calcular 24 / 8 ^ (div), donde "div" es el divisor (3, 5, 7,..)
; del termino. Para ello hay que obtener la posicion de los bits correspon-
; dientes a "xorear" de la siguiente manera:
;
;          24 / 8 ^ (div) = 3 / 8 ^ (div - 1)
;                         = 3 / 2 ^ (3 * div - 3)
;                         = 2 ^ -(3 * div - 4) + 2 ^ -(3 * div - 3)
;
; Por lo tanto los bits a "xorear" son los correspondientes a las potencias
; negativas de dos: 3 * div - 4 y 3 * div - 3.
;
; Ahora bien, como solo trabajamos con la parte decimal de los numeros, los
; bits se numeran como bit 0 -> 2^(-1), bit 1 -> 2^(-2), etc., las posiciones
; como numero de bit seran 3 * div - 5 y 3 * div - 4.
;
        mov edi,divisor             ;EDI <- div
        lea edi,[edi*2+edi]         ;EDI <- 3 * div
        sub edi,5                   ;EDI <- 3 * div - 5
        jb short OtrasPotencias     ;Si hay carry es porque el divisor
                                    ;es 1. En este caso 24 / 8^div es un
                                    ;numero entero, por lo que no hay que
                                    ;hacer nada.
        mov cx,di                   ;Bits 4-0 de CL: Posicion de bit.
        mov eax,0C0000000h          ;Inicializar para rotacion.
        ror eax,cl                  ;Realizar rotacion.
        shr edi,3                   ;Convertir numero de bit a numero de byte.
        and di,0FFFCh               ;Alinear a DWord.
        mov ecx,CantDWords          ;ECX = Cantidad de DWords del numero.
        xor es:[edi+ecx*4],eax      ;Realizar "xor" para poner o sacar potencia
OtrasPotencias:
        mov eax,5                   ;Poner o sacar potencia de (1/32).
        call Potencia
        mov eax,7                   ;Poner o sacar potencia de (1/128).
Potencia:
        mul divisor
        sub eax,2
CambiarBit:
        mov cl,al
        inc edx
        ror edx,cl
        shr eax,5
        mov ecx,CantDWords
        cmp eax,ecx
        lea eax,[eax*4]
        jae short FinSubrutina
        xor es:[eax+ecx*4],edx
FinSubrutina:
        ret
;-----------------------------------------------------------------------------
NuevoTermino:
        call PunteroSubrPotenciasDe2   ;Retirar potencias de 2 del
                                       ;termino anterior.
        mov edi,ecx                    ;EDI = Cantidad de DWords del numero.
        shl edi,2                      ;EDI = Puntero al termino.
        mov eax,PrimDWord              ;EAX = Numero de palabra mas
                                       ;significativa distinta de ceros.
        sub ecx,eax                    ;ECX = Cantidad de palabras a procesar.
        jbe short CantidadPalabrasDelTerminoOk    ;Saltar si es <= 0.
        lea edi,[edi+eax*4]            ;EDI = Puntero a la palabra mas
                                       ;significativa distinta de ceros.
        mov esi,Denominador            ;ESI = Denominador del arc tg/arg tgh.
        xor edx,edx                    ;EDX = Resto de la primera division
                                       ;por el denominador.
        xor ebx,ebx                    ;EBX = Resto de la segunda division
                                       ;por el denominador.
        cld                            ;Procesar en direcciones ascendentes
                                       ;(para realizar una division se deben
                                       ;tomar primero los Dwords mas signif.)
        push ecx                       ;Preservar la cantidad de palabras a
                                       ;procesar.
CicloDivPorDenominadorALa2:
        mov eax,es:[edi]               ;Obtener el DWord del dividendo.
        div esi                        ;Realizar la primera division.
        xchg ebx,edx                   ;EDX = Resto de la segunda division
                                       ;anterior.
                                       ;EBX = Resto de la primera division
                                       ;actual.
        div esi                        ;Realizar la segunda division.
        xchg ebx,edx                   ;EDX = Resto de la primera division
                                       ;actual.
                                       ;EBX = Resto de la segunda division
                                       ;actual.
        stos dword ptr es:[edi]        ;Almacenar el cociente.
        eloop CicloDivPorDenominadorALa2  ;Cerrar el ciclo.
        pop ecx                        ;Restaurar la cantidad de palabras a
                                       ;procesar.
        std                            ;Procesar en direcciones descendentes
                                       ;(para realizar una multiplicacion se
                                       ;deben tomar primero los Dwords menos
                                       ;significativos).
        sub edi,4                      ;Apuntar al DWord menos significativo.
        mov esi,Numerador              ;ESI = Multiplicador.
        mov ebx,esi                    ;EBX = Multiplicador / 2 (para redondeo)
        shr ebx,1
CicloMultPorNumerador:
        mov eax,es:[edi]               ;Obtener DWord del factor.
        mul esi                        ;Realizar la multiplicacion.
        add eax,ebx                    ;Sumar acarreo de multiplic. anterior.
        adc edx,0                      ;Ajustar el DWord mas significativo.
        mov ebx,edx                    ;Guardar acarreo de la multiplicacion.
        stos dword ptr es:[edi]        ;Almacenar DWord del producto.
        eloop CicloMultPorNumerador    ;Cerrar ciclo.
        and eax,eax                    ;El DWord mas significativo vale cero?
        jnz short CantidadPalabrasDelTerminoOk ;Saltar si no es asi.
        inc PrimDWord                  ;El primer DWord es el siguiente.
CantidadPalabrasDelTerminoOk:
        add divisor,2                  ;El divisor 3,5,7,... de la formula
                                       ;debe ser incrementado.
        call PunteroSubrPotenciasDe2   ;Poner las potencias binarias en
                                       ;el termino.
        mov esi,edi                    ;Posicion dentro del termino de la DWord
                                       ;mas significativa.
        mov edi,ecx                    ;EDI = Cantidad de DWords del numero.
        shl edi,3                      ;Apuntar al principio del cociente.
        add edi,esi                    ;Apuntar al DWord mas significativo
                                       ;del cociente.
        shr esi,2                      ;Convertir a cantidad de DWords.
        sub ecx,esi
        jbe short BinarioADecimalNoRET
        xor edx,edx                    ;Inicializar resto de la division.
        mov ebx,divisor                ;Obtener divisor.
        cld                            ;Procesar en direcciones ascendentes
                                       ;(para realizar una division se deben
                                       ;tomar primero los DWords mas signif.)
        add esi,CantDWords             ;
        shl esi,2                      ;Convertir a cantidad de bytes.
        push ecx
CicloDivisionPorValorVariable:
        lods dword ptr es:[esi]
        div ebx
        stos dword ptr es:[edi]
        eloop CicloDivisionPorValorVariable
        pop ecx
        mov edi,CantDWords
        shl edi,2                      ;Convertir a bytes.
        lea esi,[edi*2+edi-4]          ;Apuntar al DWord menos significativo
                                       ;del termino.
        add edi,-4                     ;Apuntar al DWord menos significativo
                                       ;de ls suma.
        std
        ret
;----------------------------------------------------------------------------
BinarioADecimalNoRET:
        pop ax                         ;No volver a rutina llamadora.
BinarioADecimal:
        mov al,bufferdos               ;Obtener la parte entera (en ASCII).
        mov BufferNumeroFormateado,al  ;Almacenar el valor en caso de /f.
        mov ax,4400h                   ;Llamada IOCTL.
        mov bx,HANDLE_STDOUT           ;Handle STDOUT.
        call dos
        test dl,dl                     ;Bit 7 de DL = 1: Dispositivo.
                                       ;            = 0: Archivo.
        jns short TextoFormateado?     ;Saltar si se redirecciono STDOUT.
        mov MostrarEnInt8,0            ;No mostrar texto en INT 8.
        mov dx,offset TiempoTotal
        mov cx,2
        mov bx,cx                      ;BX = Handle STDERR.
        mov ah,40h
        call dos
TextoFormateado?:
        test Indicadores,Formato       ;El texto debe estar formateado?
        jnz short ComenzarConversion   ;Saltar si es asi.
        mov dx,offset bufferdos
        mov cx,2                       ;Mostrar parte entera y coma.
        mov bx,handle
        mov ah,40h
        call dos
;Convertir de binario a decimal.
ComenzarConversion:
        std
        mov eax,CantDWords
        lea eax,[eax*4 - 4]
        mov UltDWord,eax
MultiplicarPorMaxPotencia:
        mov ecx,CantDWords
        mov edi,UltDWord
        mov esi,MaxPotencia             ;Cargar maxima potencia en un DWord.
        xor ebp,ebp                     ;Carry primera multiplicacion.
        xor ebx,ebx                     ;Carry segunda multiplicacion.
CicloMultPorMaxPotencia:
        mov eax,es:[edi]                ;Obtener DWord del numero.
        mul esi                         ;Realizar la primera multiplicacion.
        add eax,ebp
        adc edx,0
        mov ebp,edx
        mul esi                         ;Realizar la segunda multiplicacion.
        add eax,ebx
        adc edx,0
        mov ebx,edx
        stos dword ptr es:[esi]
        eloop CicloMultPorMaxPotencia
;En este momento EBP tiene los digitos mas significativos y EBX los siguientes.
FinMultEnConv:
        mov eax,LogMaxPotencia          ;Corresponde a primera multiplicacion.
        add SumaLogPotencia,eax
        jnc short SegundaSumaLogs
        sub UltDWord,4
        dec CantDWords
SegundaSumaLogs:
        add SumaLogPotencia,eax         ;Corresponde a segunda multiplicacion.
        jnc short ObtenerSiguientesDigitos
        sub UltDWord,4
        dec CantDWords
ObtenerSiguientesDigitos:
        mov di,offset bufferdos[-1] ;DI = Puntero al principio del buffer
                                ;temporario de digitos en pantalla.
        mov cx,word ptr Exponente ;Cantidad de digitos del numero.
        add di,cx
        add di,cx               ;DI = Puntero al ultimo digito.
        movzx esi,Base          ;ESI = Base.
        call EBX_A_OtraBase
        mov cx,word ptr Exponente ;Cantidad de digitos del numero.
        mov ebx,ebp             ;Convertir los digitos mas significativos.
        call EBX_A_OtraBase
        mov ecx,digitos
        mov eax,Exponente
        add eax,eax
        cmp ecx,eax
        jbe short CantidadDeDigitosAMostrarOK
        mov ecx,eax
CantidadDeDigitosAMostrarOK:
        test Indicadores,Formato       ;Hay que formatear el texto?
        jnz short FormatearTexto       ;Saltar si es asi.
        jecxz FinConversion            ;Saltar si se indico cero decimales.
        mov bx,handle
        mov dx,offset bufferdos
        mov ah,40h
        call dos
        mov eax,Exponente
        shl eax,1
        sub digitos,eax
OtrosDigitos:
        jbe short FinConversion
        jmp MultiplicarPorMaxPotencia
FormatearTexto:
        jecxz UltimaLineaTextFormateado
        call DigitosFormateados
        cmp Digitos,0
        jnz OtrosDigitos
UltimaLineaTextFormateado:
        call MostrarBufferEnPantalla
FinConversion:
Final:
        mov dx,OldInt8Offset     ;Poner nuevamente el viejo vector de
        mov cx,OldInt8Segment    ;interrupcion 8 en modo real.
        mov ax,0201h
        mov bl,8
        int 31h
        mov dx,offset TiempoTotal ;CR - LF
        mov cx,2
        mov bx,cx              ;Handle STDERR.
        mov ah,40h
        call dos
        mov dx,offset TiempoTotal
        mov cx,LongitudTiempoTotal
        mov bx,HANDLE_STDERR   ;Handle STDERR.
        mov ah,40h
        call dos
        mov ax,0007h           ;Cambiar el descriptor para que apunte al buffer
        mov bx,Selector        ;de video.
        mov cx,000Bh
        mov dx,8000h
        int 31h
        call obtencur          ;Obtener (en DI) la posicion del cursor.
        mov es,Selector
        call MostrarTiempo     ;Mostrar el tiempo total de ejecucion.
        mov dx,offset TiempoTotal ;CR - LF
        mov cx,2
        mov bx,cx              ;Handle STDERR.
        mov ah,40h
        call dos
        mov bx,Selector        ;Liberar el selector pedido al servidor de DPMI
        mov ax,0001h
        int 31h
        mov si,HandleMemoriaAlto ;Liberar la zona de memoria pedida.
        mov di,HandleMemoriaBajo
        mov ax,0502h
        int 31h
        mov ax,4c00h           ;Fin del programa.
        int 21h
;----------------------------------------------------------------------------
; Subrutina dos: Sirve para llamar al DOS en modo real.
;
dos:
        mov word ptr ModoReal.reg_EAX,ax  ;Poner los registros en la
        mov word ptr ModoReal.reg_EBX,bx  ;estructura para modo real.
        mov word ptr ModoReal.reg_ECX,cx
        mov word ptr ModoReal.reg_EDX,dx
        mov word ptr ModoReal.reg_ESI,si
        mov word ptr ModoReal.reg_EDI,di
        mov word ptr ModoReal.reg_SS,0
        mov word ptr ModoReal.reg_SP,0
        mov ax,0300h                      ;Llamar a interrupcion en modo real.
        mov bx,0021h                      ;DOS = Interrupcion 21h.
        xor cx,cx                         ;Cantidad de bytes a pasar en stack:
                                          ;cero.
        mov di,offset ModoReal            ;Apuntar a la estructura de modo real
        movzx edi,di
        push ds
        pop es
        int 31h
        mov ah,byte ptr ModoReal.reg_Flags
        sahf
        mov es,Selector
        mov ax,word ptr ModoReal.reg_EAX
        mov bx,word ptr ModoReal.reg_EBX
        mov cx,word ptr ModoReal.reg_ECX
        mov dx,word ptr ModoReal.reg_EDX
        mov si,word ptr ModoReal.reg_ESI
        mov di,word ptr ModoReal.reg_EDI
        ret
;----------------------------------------------------------------------------
; Manejador de la interrupcion 8 (timer). Debe mostrar la cantidad de digitos
; y el tiempo transcurrido.
;
Int8Handler:
        push ds                        ;Preservar los registros utilizados.
        push es
        push esi
        push di
        push bx
        push eax
        push edx
        push cx
        push cs                        ;Cargar el registro de segmento de
        pop ds                         ;datos.
        cld
        mov di,offset FraccSegundo
        add dword ptr [di],235903917   ;2^32 / 18,20648193
        jnc short CantDigitosAPantalla
        dec di                         ;Apuntar a los segundos.
        inc byte ptr [di]              ;Incrementar segundos.
        cmp byte ptr [di],60           ;Se termino el minuto?
        jnz short TiempoActualizado    ;Saltar si no es asi.
        mov byte ptr [di],0
        dec di                         ;Apuntar a los minutos.
        inc byte ptr [di]              ;Incrementar minutos.
        cmp byte ptr [di],60           ;Se termino la hora?
        jnz short TiempoActualizado    ;Saltar si no es asi.
        mov byte ptr [di],0
        dec di                         ;Apuntar a las horas.
        inc byte ptr [di]              ;Incrementar horas.
        cmp byte ptr [di],24           ;Se termino el dia?
        jnz short TiempoActualizado    ;Saltar si no es asi.
        mov byte ptr [di],0
        dec di                         ;Apuntar a dias.
        inc byte ptr [di]              ;Incrementar horas.
        cmp byte ptr [di],100          ;Pasaron cien dias?
        jnz short TiempoActualizado    ;Saltar si no es asi.
        mov byte ptr [di],0
TiempoActualizado:
        cmp MostrarEnInt8,0            ;Hay que mostrar numero y tiempo?
        jz FinInter8                   ;Salir de la interrupcion si no es asi.
        les di,CursorVideo             ;Obtener la posicion del cursor.
        add di,15*2                    ;Apuntar al tiempo.
        call MostrarTiempo             ;Mostrar tiempo transcurrido.
CantDigitosAPantalla:
        cmp MostrarEnInt8,0            ;Hay que mostrar numero y tiempo?
        jz FinInter8                   ;Salir de la interrupcion si no es asi.
        les di,CursorVideo             ;Obtener la posicion del cursor.
        mov si,PunteroBase
        xor eax,eax
        test Indicadores,E             ;Se esta calculando el numero e?
        jz short NoSeCalculaE
        mov edx,67108864               ;2^64 / 64
        div dword ptr [si]
        mul PrimDWord
        jmp short AjustarDigitos
NoSeCalculaE:
        mov edx,6291456                ;2^64 * log 8 / log (2^32) / 64.
        div dword ptr [si]             ;2^32 * log 8 / log Base
        mul Divisor                    ;EDX = Cantidad de digitos / 64.
AjustarDigitos:
        shld edx,eax,6                 ;EDX = Cantidad de digitos.
        mov bl,0
        mov esi,1000000000
        call Div1Digito
        mov esi,100000000
        call Div1Digito
        mov esi,10000000
        call Div1Digito
        mov esi,1000000
        call Div1Digito
        mov esi,100000
        call Div1Digito
        mov esi,10000
        call Div1Digito
        mov si,1000
        call Div1Digito
        mov si,100
        call Div1Digito
        mov si,10
        call Div1Digito
        add dl,"0"
        mov es:[di],dl
FinInter8:
        pop cx                  ;Restaurar los registros afectados.
        pop edx
        pop eax
        pop bx
        pop di
        pop esi
        pop es
        pop ds
        jmp cs:OldInt8Handler   ;Ir al viejo manejador de INT 8.
;---------------------------------------------------------------------------
PonerVectorInterrupcion:
        mov dx,offset TextoSegundaLinea
        mov cx,LongitudTextoSegundaLinea
        mov bx,HANDLE_STDERR    ;Handle STDERR.
        mov ah,40h
        call dos                ;Mostrar el texto.
        call obtencur           ;Obtener la direccion del cursor.
        mov CursorVideoOffset,di
        mov ax,0200h            ;Obtener el vector de interrupcion (modo real).
        mov bl,8                ;Numero de interrupcion.
        int 31h
        mov OldInt8Offset,dx    ;Preservar el vector.
        mov OldInt8Segment,cx
        mov ax,0201h            ;Setear el vector de interrupcion (modo real).
        mov bl,8                ;Numero de interrupcion.
        mov cx,ModoReal.reg_DS
        mov dx,offset Int8Handler
        int 31h
        ret                     ;Fin de la subrutina.
;---------------------------------------------------------------------------
; Subrutina para obtener la posicin del cursor en la pantalla.
; Salida: DI = Posicion en buffer de video que corresponde al cursor.
;
obtencur:mov dx,port6845     ;DX = Port de control del controlador de video.
        mov al,0Eh           ;Indicarle que hay que leer la parte alta
        out dx,al            ;(MSB) de la posicin del cursor.
        jmp $+2
        jmp $+2
        inc dx               ;Apuntar al port de datos del 6845.
        in al,dx             ;Leer el valor.
        mov ah,al            ;Ponerlo en la parte alta de AX.
        jmp $+2
        jmp $+2
        dec dx               ;Apuntar al port de control del 6845,
        mov al,0Fh           ;Indicarle que hay que leer la parte baja
        out dx,al            ;(LSB) de la posicin del cursor.
        jmp $+2
        jmp $+2
        inc dx               ;Apuntar al port de datos del 6845.
        in al,dx             ;Leer el valor.
        add ax,ax
        mov di,ax            ;Poner el resultado en DI.
        ret                  ;Fin de la subrutina.
;----------------------------------------------------------------------------
; Subrutina MostrarTiempo: Mostrar el tiempo transcurrido.
; Entrada: ES:DI = Posicion en buffer de video.
;
MostrarTiempo:
        mov si,offset Dias
        mov cx,4
        cld
CicloMostrarTiempo:
        lodsb
        aam
        add ax,"00"
        mov es:[di],ah
        mov es:[di+2],al
        mov al,[si-5]
        mov es:[di+4],al
        mov byte ptr es:[di+6]," "
        add di,8
        loop CicloMostrarTiempo
        ret
;-----------------------------------------------------------------------------
Div1Digito:
        mov eax,edx
        xor edx,edx
        div esi
        add al,"0"
        cmp al,"0"
        jnz short MostrarElDigito
        test bl,bl
        jz short ApuntarSiguienteDigito
MostrarElDigito:
        mov bl,1
        stosb
        inc di
ApuntarSiguienteDigito:
        ret
;-----------------------------------------------------------------------------
; Subrutina DigitosFormateados: Halla un digito del numero y, si esta
; habilitado con el switch /F, realiza el formateo de texto.
;
; Entrada: bufferdos: Digitos en ASCII.
;          CX = Cantidad de digitos a mostrar.
;          DI = Puntero al buffer donde se almacenan los digitos en ASCII.
;
DigitosFormateados:
        push es          ;Preservar puntero al numero en binario.
        push ds
        pop es
        cld
        mov si,offset bufferdos
        mov dx,si
        add dx,word ptr Exponente
        add dx,word ptr Exponente
        mov ah,0         ;Mostrar los ceros mas significativos como espacios.
        test Indicadores,MostrarCantDigitos
        jnz short CicloMostrarSinFormatear
        test Indicadores,Formato
        jnz short MostrarFormateado
        mov ah,1         ;Mostrar los ceros mas significativos como ceros.
CicloMostrarSinFormatear:
        lodsb
        cmp al,"0"
        jnz short DigitoSignificativo
        test ah,ah
        jnz short DigitoSignificativo
        cmp si,dx
        jz short MostrarDigito
        mov al," "
        jmp short MostrarDigito
DigitoSignificativo:
        mov ah,1
MostrarDigito:
        stosb
        loop CicloMostrarSinFormatear
        std
        pop es
        ret
MostrarFormateado:
        dec digitos
        mov di,PosicionDigitoEnBuffer
        mov bx,InformacionDeFormateado  ;BH = Numero de grupo (0-9).
                                        ;BL = Digito dentro del grupo (1-5).
        inc bx                          ;Indicar siguiente digito del grupo.
        cmp bl,6                        ;Fin del grupo?
        jnz short PonerDigitoFormateado ;Saltar si no es asi.
        mov bl,1                        ;Indicar inicio del grupo.
        inc bh                          ;Indicar siguiente grupo.
        cmp bh,10                       ;Fin de la linea?
        jnz short PonerEspacio          ;Ir a mostrar espacio si no es asi.
        push si
        call MostrarBufferEnPantalla
        pop si
        mov di,offset BufferNumeroFormateado
                                ;Borrar el buffer de digitos formateados.
        push cx                 ;Preservar contador de digitos a mostrar.
        cld
        mov cx,75
        mov al," "
        rep stosb
        pop cx                  ;Restaurar contador de digitos a mostrar.
        mov bx,HANDLE_STDOUT
        mov di,offset BufferNumeroFormateado + 1
                                ;Apuntar al segundo byte del buffer.
PonerEspacio:
        inc di                  ;Apuntar al siguiente caracter.
PonerDigitoFormateado:
        movsb           ;Copiar el digito en buffer de digitos formateados.
        mov PosicionDigitoEnBuffer,di
        mov InformacionDeFormateado,bx
        inc DigitosMostrados    ;Indicar que se mostro un digito mas.
        loop MostrarFormateado
        std
        pop es
        ret                     ;Fin de la subrutina.
;---------------------------------------------------------------------------
MostrarBufferEnPantalla:
        push cx
        or Indicadores,MostrarCantDigitos
        push esi
        mov ebx,DigitosMostrados
        mov di,offset BufferNumeroFormateado[74]
                                ;DI = Puntero al final del buffer
                                ;     del numero a mostrar a la derecha.
        mov esi,10              ;ESI = Base.
        mov cx,13               ;CX = Cantidad de digitos a convertir.
        call EBX_A_OtraBase
        mov cx,12
CicloEliminarCerosIzq:
        inc di
        cmp byte ptr [di],"0"
        jnz short CerosEliminados
        mov byte ptr [di]," "
        loop CicloEliminarCerosIzq
CerosEliminados:
        pop esi
        and Indicadores,not MostrarCantDigitos
        mov bx,handle
        mov dx,offset BufferNumeroFormateado
        mov cx,79
        push es
        mov ah,40h
        call dos
        pop es
        pop cx
        ret
;------------------------------------------------------------------------------
; Conversion binario (EBX = 32 bits) a otra base.
;
; Entrada: EBX = Doble palabra a convertir.
;          ESI = Base.
;          CX = Cantidad de digitos a convertir.
; Salida:  bufferdos: Numero en ASCII.
;
EBX_A_OtraBase:
bucle_bin_a_dec:
        mov eax,ebx             ;Cargar el dividendo.
        xor edx,edx             ;Extender el dividendo.
        div esi                 ;Realizar la division.
        mov ebx,eax             ;Salvar el cociente, que sera el dividendo.
        add dl,"0"              ;Convertir a ASCII.
        cmp dl,"9"
        jbe short PonerDigito
        add dl,"A" - ("9" + 1)
PonerDigito:
        mov [di],dl             ;Guardar el resto en el buffer temporario
                                ;(de esta manera se obtiene el digito
                                ;"de la derecha" o menos significativo).
        dec di                  ;Apuntar al digito anterior.
        loop bucle_bin_a_dec    ;Saltar si no se acabo el buffer.
        ret                     ;Fin de la subrutina.
;---------------------------------------------------------------------------
; Subrutina switch: Procesa los switches de linea de comandos inicializando
; los flags correspondientes.
;
; Entrada: BX = Puntero a la linea de comandos.
;
switch: cmp byte ptr [bx],"/"            ;Se encontro el caracter de switch?
        jz short SeEncontroSwitch        ;Saltar si es asi.
        cmp byte ptr [bx],ASCII_CR       ;Fin de la linea de comandos?
        jz short FinSwitch               ;No hay switches si es asi.
        inc bx                           ;Apuntar al sig. caracter de la linea.
        cmp byte ptr [bx-1]," "          ;Habia un espacio o caracter tab?
        jbe switch                       ;Saltearlo si es asi.
        lea dx,[bx-1]                    ;DX = Puntero al nombre del archivo.
buscar_fin_de_nombre_de_archivo:
        inc bx                           ;Apuntar al siguiente caracter.
        cmp byte ptr [bx-1]," "          ;Fin del nombre de archivo?
        ja buscar_fin_de_nombre_de_archivo ;Seguir buscando si no es asi.
        mov [bx-1],bh                    ;Convertir el nombre a ASCIIZ.
        mov ah,3Ch                       ;Intentar crear el archivo.
        xor cx,cx
        call dos
        jc short JmpToError              ;Saltar si no se pudo.
        mov handle,ax                    ;Almacenar el handle.
FinSwitch:
        ret                              ;Fin de la subrutina.
JmpToError:
        jmp error                        ;Ir a la rutina de error.
SeEncontroSwitch:
        mov ax,[bx+1]                    ;Obtener el tipo de switch.
        and ax,0DFDFh                    ;Convertir a mayusculas.
        cmp al,"E"                       ;Se desea calcular el valor de e?
        jnz short EsLN2?                 ;Saltar si no es asi.
        or Indicadores,E                 ;Indicar que se desea calcular e.
        inc bx                           ;Saltear el switch.
        inc bx
        jmp switch                       ;Ir a verificar si hay mas switches.
EsLN2?:
        cmp ax,"NL"                      ;Es la primera parte de LN 2?
        jnz short EsFormato?             ;Saltar si no es asi.
        cmp byte ptr [bx+3],"2"          ;Es la segunda parte de LN 2?
        jnz short EsFormato?             ;Saltar si no es asi.
        or Indicadores,LN2               ;Indicar que se desea calcular ln 2.
        add bx,4                         ;Saltear el switch.
        jmp switch                       ;Ir a verificar si hay mas switches.
EsFormato?:
        cmp al,"F"                       ;Es el switch de texto formateado?
        jnz short EsBase?                ;Saltar si no es.
        or Indicadores,Formato           ;Indicar que el texto debe formatearse
        add bx,2                         ;Saltear el switch.
        jmp switch                       ;Ir a verificar si hay mas switches.
EsBase?:
        mov ax,[bx+1]
        and al,0DFh                      ;Convertir a mayusculas el primer
                                         ;caracter.
        cmp ax,":B"                      ;Es base?
        jnz JmpToError                   ;Ir a error si el switch es invalido.
        mov cl,0                         ;Inicializar la base.
        add bx,3                         ;Apuntar a los digitos.
CicloBase:
        mov ch,[bx]                      ;Obtener el byte de la linea.
        sub ch,"0"                       ;Convertir de ASCII a binario.
        cmp ch,9                         ;Es un digito?
        ja short FinBase                 ;Fin de busqueda de digitos si no es.
        mov al,10                        ;Multiplicar la base por 10.
        mul cl
        jc JmpToError                    ;Error si la base supera 256.
        add al,ch                        ;Obtener la nueva base.
        cmp al,36                        ;La base es mayor que 36?
        ja JmpToError                    ;Error si la base es mayor que 36.
        mov cl,al                        ;Preservar el valor.
        inc bx                           ;Apuntar al siguiente byte.
        jmp CicloBase                    ;Volver a pedir mas digitos.
FinBase:
        cmp cl,2                         ;La base es menor que 2?
        jb JmpToError                    ;Error si es asi.
        mov Base,cl                      ;Almacenar la base.
        jmp switch
;-----------------------------------------------------------------------------
fin_codigo:
codigo  ends
        end comienzo
